/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef TX_ELL_MATRIX_H
#define TX_ELL_MATRIX_H

#include <cusp/ell_matrix.h>
#include <tx_common.h>

// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_ell_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2012 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------

#define CUSPMATRIXELL cusp::ell_matrix<PetscInt,PetscScalar,cusp::device_memory>

template <typename index_type, typename value_type, unsigned int BLOCK_SIZE, class OPERATOR>
__launch_bounds__(BLOCK_SIZE,1)
__global__ void spmv_ell_kernel(const index_type num_rows, 
				const index_type num_cols, 
				const index_type num_cols_per_row,
				const index_type pitch,
				const index_type * Aj,
				const value_type * Ax, 
				const value_type * x, 
				value_type * y,
				OPERATOR OP)
{
    const index_type invalid_index = cusp::ell_matrix<index_type, value_type, cusp::device_memory>::invalid_index;

    const index_type thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const index_type grid_size = gridDim.x * blockDim.x;

    for(index_type row = thread_id; row < num_rows; row += grid_size)
    {
        value_type sum = 0;

        index_type offset = row;

        for(index_type n = 0; n < num_cols_per_row; n++)
        {
            const index_type col = Aj[offset];

            if (col != invalid_index)
            {
                const value_type A_ij = Ax[offset];
                sum += A_ij * x[col];
                //sum += A_ij * fetch_x<UseCache>(col, x);
            }

            offset += pitch;
        }
	y[row] = OP(y[row],sum);
    }
}


template <typename index_type, typename value_type, class OPERATOR>
void spmv_ell(const cudaStream_t stream,
	      const CUSPMATRIXELL& A,
	      const value_type* x, 
	      value_type* y,
	      OPERATOR OP)
{
  const size_t BLOCK_SIZE = 256;
  const size_t MAX_BLOCKS = cusp::detail::device::arch::max_active_blocks(spmv_ell_kernel<index_type,value_type,BLOCK_SIZE, OPERATOR>, BLOCK_SIZE, (size_t) 0);
  const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, DIVIDE_INTO(A.num_rows, BLOCK_SIZE));
  
  const index_type pitch               = A.column_indices.pitch;
  const index_type num_entries_per_row = A.column_indices.num_cols;
  
  // TODO generalize this
  assert(A.column_indices.pitch == A.values.pitch);
  
  //if (UseCache)
  //  bind_x(x);
  
  spmv_ell_kernel<index_type,value_type,BLOCK_SIZE> <<<NUM_BLOCKS, BLOCK_SIZE, 0, stream>>>
    (A.num_rows, A.num_cols,
     num_entries_per_row, pitch,
     thrust::raw_pointer_cast(&A.column_indices.values[0]), 
     thrust::raw_pointer_cast(&A.values.values[0]),
     x, y, OP);
  
  //if (UseCache)
  //  unbind_x(x);
}


/***********************/
/*      Ell Matrix     */
/***********************/
template <typename index_type, typename value_type>
class EllMatrix {
public:
  typedef CUSPMATRIXELL MatrixType;

  /* Constructor */
  EllMatrix() {}

  /* Destructor */
  virtual ~EllMatrix() {
  }

  /* Set the Matrix explicitly on the device*/
  void setMatrix(const cusp::csr_matrix<index_type,value_type,cusp::host_memory>& matCPU) {
    A = matCPU;
  }  

  /* Set the Matrix */
  void setMatrix(const PetscInt numRows, const PetscInt numCols, const PetscInt nz,
		 const PetscInt * row_offsets, const PetscInt * column_indices,
		 const PetscScalar * values) {

    // create a temporary 
    cusp::csr_matrix<PetscInt,PetscScalar,cusp::host_memory> B(numRows,numCols,nz);    
    B.row_offsets.assign(row_offsets, row_offsets+numRows+1);
    B.column_indices.assign(column_indices, column_indices+nz);
    B.values.assign(values, values+nz);
    // Let CUSP do the conversion
    A = B;
  }

  /* Set the Indices For cprow==true (compressed row) */
  void setCPRowIndices(const index_type * indices, const index_type m) {
    // does nothing as of now
  }

  /* Set the inode data for when inode.use==true (inodes) */
  virtual void setInodeData(const PetscInt * inodesCPU, const int inodeCountPlusOne, const int nnzPerRowMaxCPU,
                            const int nodeMaxCPU, const int nodeCountCPU) {
    // does nothing as of now
  }

  /* SpMV */
  int multiply(const cudaStream_t stream, const value_type *x, value_type *y) {
    spmv_ell<index_type, value_type>(stream, A, x, y, Equals<value_type>());
    return (int) cudaGetLastError();
  }

  /* SpMV Add*/
  int multiplyAdd(const cudaStream_t stream, const value_type *x, value_type *y) {
    spmv_ell<index_type, value_type>(stream, A, x, y, PlusEquals<value_type>());
    return (int) cudaGetLastError();
  }


private:
  MatrixType A;/* container for the matrix */
};

#endif // TX_ELL_MATRIX_H
